/*
 * DB: Global Script Support
 * Copyright (C) 2002 by Scott Wallace (swallace@eecs.umich.edu)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <PalmOS.h>
#ifdef CONFIG_HANDERA
#include <Vga.h>
#endif
#include "enum.h"
#include "db.h"
#include "calcvalue.h"
#ifdef CONFIG_HANDERA
#include <Vga.h>
#include "util.h"
#endif


#define GetObjectPtr(f,i) FrmGetObjectPtr((f),FrmGetObjectIndex((f),(i)))

#define MAX_GLB_SCRIPT_INDEX 50

extern UInt16 FeatureEditor_FeatureNum;


Boolean
GlbScriptEditorHandleEvent( EventPtr event ) {
	FormPtr form;
	FieldPtr fld;

	switch (event->eType) {

    case frmOpenEvent:
	{
		DataSourceFeature_GlobalScript *orig_gsd;
		MemHandle h;

        /* Retrieve the current form. */
        form = FrmGetActiveForm();
#ifdef CONFIG_HANDERA
		if (FeatureSet.Handera) {
			VgaFormModify(form, vgaFormModify160To240);
			if (!UtilCheckSizeForm(form, 0)) {
				FrmAlert(alertID_FormTooBig);
				FrmReturnToForm(0);
				return true;
			}
		}
#endif
			orig_gsd = CurrentSource->ops.GetFeature( CurrentSource, featureGlobalScript, FeatureEditor_FeatureNum );

			
			if ( orig_gsd != NULL ) {
				orig_gsd->c.scriptText = CurrentSource->driver->class_ops.DecompileScript( &(orig_gsd->c) );
				

				fld = GetObjectPtr( form, ctlID_GlobalScriptEdit_ScriptName );
				h = FldGetTextHandle( fld );
				
				if ( !h ) {
					h = AllocateMemHandle( 32 ); 
					RecordMemInfo( h, "freed by system (but not reported!) when text field is destroyed" );
					
				}

				StrNCopy( MemHandleLock( h ), orig_gsd->name, 32 );
				MemHandleUnlock( h );
				FldSetTextHandle( fld, h );
				

				fld = GetObjectPtr( form, ctlID_GlobalScriptEdit_ScriptField );
				h = FldGetTextHandle( fld );
				if ( h ) DeallocateMemHandle( h );
				FldSetTextHandle( fld, orig_gsd->c.scriptText );
				RecordMemInfo( orig_gsd->c.scriptText, "freed by system (but not reported!) when text field is destroyed." );
				
				
				// can't use PutFeature since we are sill using the scripText handle
				

				if ( orig_gsd->c.bytecode ) 
					DeallocateMemHandle( orig_gsd->c.bytecode );
				

				DeallocateMemPtr( orig_gsd );
			}

			
			FrmDrawForm( form );
			return true;
	}		
			break;

	case frmCloseEvent:
		break;

	case ctlSelectEvent:
	{
		DataSourceFeature_GlobalScript gsd;
		MemHandle h;
		Boolean invalid;

		form = FrmGetActiveForm();

		if ( event->data.ctlSelect.controlID == ctlID_GlobalScriptEdit_OK ) {
			ScriptContext sc;

			sc.src = CurrentSource;
			sc.context = LIST_VIEW_INIT_CONTEXT;
			sc.flags = ScriptContext_DataAccess_RTC;
			sc.rec_field = 0;
				
			invalid = false;

            if (! (CurrentSource->ops.SupportsFeature(CurrentSource,
													  featureGlobalScript) &
				   FEATURE_SUPPORT_EDIT) ) {
				
				FrmReturnToForm( 0 );
				return true;
            }
			
			fld = GetObjectPtr( form, ctlID_GlobalScriptEdit_ScriptName );
			h = FldGetTextHandle( fld );
			StrNCopy( gsd.name, MemHandleLock( h ), 32 );
			MemHandleUnlock( h );
			if ( StrLen( gsd.name ) == 0 ){
				break;
			}
				
			fld = GetObjectPtr( form, ctlID_GlobalScriptEdit_ScriptField );
			gsd.c.scriptText = FldGetTextHandle( fld );
			gsd.c.bytecode = NULL;
			
			if ( CurrentSource->driver->class_ops.CompileScript( &sc, &(gsd.c), false) ) {
				FldSetTextHandle( fld, NULL );
				CurrentSource->ops.StoreFeature( CurrentSource, featureGlobalScript, FeatureEditor_FeatureNum, &gsd );

				FrmUpdateForm( formID_NamesEditor, updateChangeForm );
				DeallocateMemHandle( gsd.c.bytecode );
				FrmReturnToForm( 0 );
				return true;
			}
			
		}
		else {
			// Cancel
			
			FrmReturnToForm( 0 );
			return true;
			
		}
	}
	break;
	case keyDownEvent:
		/**
		 *  Active Parenthesis
		 */
		if ( event->data.keyDown.chr == ')' ) {
			Int16 q,p,l;
			UInt32 t;
			char *txt;

            form = FrmGetActiveForm();
            fld = GetObjectPtr(form, ctlID_GlobalScriptEdit_ScriptField );
			txt = MemHandleLock(FldGetTextHandle( fld ) );
			q = StrLen( txt ) - 1;
			p = 1;
			l = FldGetInsPtPosition( fld );
			while( q > 0) {
				if ( txt[q] == '(' ) p--;
				else if ( txt[q] == ')' ) p++;
				
				if ( p < 1 ) break;
				q--;
			}
			FldSetInsPtPosition( fld, q );
			t = TimGetTicks();
			q = 0;
			while( (TimGetTicks() - t) < SysTicksPerSecond() / 3 ) {
				p += q*11;
				q++;
			}
			MemPtrUnlock( txt );
			FldSetInsPtPosition( fld, l );
			
			
		}
		break;
			
	default:
	}
	return false;

}
