
#include <PalmOS.h>
#include <TxtGlue.h>

#include "db.h"
#include "enum.h"
#include "io.h"
#include "prefs.h"
#include "listview.h"
#include "util.h"
#include "names.h"
#ifdef CONFIG_HANDERA
#include <Vga.h>
#endif

#include "localsearch.h"

#ifndef __PLUGIN__

FindData FindInfo;

#define GetObjectPtr(f,i) FrmGetObjectPtr((f),FrmGetObjectIndex((f),(i)))




Boolean
FindDialogHandleEvent(EventPtr event)
{
    static char ** names;

    FormPtr form;
    FieldPtr fld;
    ControlPtr ctl;
    ListPtr lst;
    Char * str;
    UInt16 i, objIndex;
    Int16 width, w;
    RectangleType r;
    MemHandle strH;

    switch (event->eType) {
    case frmOpenEvent:
	{

		MemHandle textH;
		char *text;

		textH = MemHandleNew( const_FindDialog_FindFieldLength );
		text = (char *)MemHandleLock( textH );
		text[0] = '\0';

        form = FrmGetActiveForm();
#ifdef CONFIG_HANDERA
        if (FeatureSet.Handera) {
            VgaFormModify(form, vgaFormModify160To240);
            if (!UtilCheckSizeForm(form, 0)) {
                FrmAlert(alertID_FormTooBig);
                FrmReturnToForm(0);
                return true;
            }
        }
#endif


		if ( FindInfo.forFilterDefinition > -1 ) {
			DataSourceFeature_Filter *filt=NULL;
			FieldPtr fld;
			MemHandle h;
			char *c;

			w = FrmGetObjectIndex( form, ctlID_FindDialog_Name );
			FrmShowObject( form, w );
			fld = FrmGetObjectPtr( form, w );
			FrmShowObject( form, FrmGetObjectIndex( form, 9000 ));			
			FrmSetTitle( form, "Configure Filter" );

			ctl = GetObjectPtr(form, ctlID_FindDialog_FindButton);
			CtlSetLabel(ctl, "Ok");
			
			filt = CurrentSource->ops.GetFeature(CurrentSource, featureFilter, 
												 FindInfo.forFilterDefinition);
			
			
			h = FldGetTextHandle( fld );

			if ( !h ) {
				h = MemHandleNew( 32 ); 
				RecordMemInfo( h, "freed by system (but not reported!) when text field is destroyed" );
			}
			c = MemHandleLock( h );
			StrCopy( c, filt->name );
			MemHandleUnlock( h );
			FldSetTextHandle( fld, h );
			
			if ( DSF_FilterType(filt) == DSF_FILTER_TYPE_SIMPLE ) {

				StrNCopy( text, filt->data.simple->matchString,
						  const_FindDialog_FindFieldLength );
				
			}
			else {
				filt->data.script->scriptText = CurrentSource->driver->class_ops.DecompileScript( filt->data.script );
				StrNCopy( text, 
						  (char *)MemHandleLock(filt->data.script->scriptText),
						  const_FindDialog_FindFieldLength );

				MemHandleUnlock( filt->data.script->scriptText );
				DeallocateMemHandle( filt->data.script->scriptText );
			}
			

			CurrentSource->ops.PutFeature(CurrentSource, featureFilter, 
										  FindInfo.forFilterDefinition, filt);
		}  // end Defining a filter
		else // Just doing a quick local search
		{
			
			if (FindInfo.type == DSF_FILTER_TYPE_SIMPLE ) {
				StrNCopy( text, FindInfo.data.simple->matchString, 
						  const_FindDialog_FindFieldLength);
				
			}
			else {
				FindInfo.data.script->scriptText = CurrentSource->driver->class_ops.DecompileScript( FindInfo.data.script );
				StrNCopy( text, (char *)MemHandleLock(FindInfo.data.script->scriptText),
						  const_FindDialog_FindFieldLength );
				MemHandleUnlock( FindInfo.data.script->scriptText );
				DeallocateMemHandle( FindInfo.data.script->scriptText );

			}


			if ( CurrentSource->ops.GetCurrentFilter( CurrentSource ) == 
				 dbTempResultsFilter ) {
				w = FrmGetObjectIndex( form, ctlID_FindDialog_OperationTrigger );
				FrmShowObject( form, w );
				w = FrmGetObjectIndex( form, 9001 );
				FrmShowObject( form, w );
				lst = GetObjectPtr(form, ctlID_FindDialog_OperationList);
				LstSetSelection( lst, 0 );
				ctl = GetObjectPtr(form, ctlID_FindDialog_OperationTrigger);
				CtlSetLabel(ctl, LstGetSelectionText( lst, 0  ) );
				
			}
		}

		text[const_FindDialog_FindFieldLength-1] = '\0';
		MemHandleUnlock( textH );
		FldSetTextHandle(GetObjectPtr(form, ctlID_FindDialog_Field), textH);
		

		ctl = GetObjectPtr(form, ctlID_FindDialog_CaseSensCheckbox);
		CtlSetValue(ctl, (FindInfo.type == DSF_FILTER_TYPE_SIMPLE && FindInfo.data.simple->caseSensitive ) ? 1:0 );

		
		ctl = GetObjectPtr(form, ctlID_FindDialog_WholeWordCheckbox);
		CtlSetValue(ctl, (FindInfo.type == DSF_FILTER_TYPE_SIMPLE && FindInfo.data.simple->wholeWord ) ? 1:0 );


		/* Allocate space for an array of string pointers for field list. */
		names = (char **) AllocateMemPtr((CurrentSource->schema.numFields + 1) * sizeof(char *));
		
		/* Duplicate the "All Fields" string. */
		names[0] = CopyStringResource(stringID_AllFields);
		
		/* Figure out the width needed for the field popup list. */
		width = FntCharsWidth(names[0], StrLen(names[0]));
		for (i = 0; i < CurrentSource->schema.numFields; i++) {
			names[1 + i] = CurrentSource->schema.fields[i].name;
			w = FntCharsWidth(names[1 + i], StrLen(names[1 + i]));
			if (w > width)
				width = w;
		}
		width += 10;
		
		/* Set the width of the field popup list. */
		objIndex = FrmGetObjectIndex(form, ctlID_FindDialog_FieldList);
		lst = FrmGetObjectPtr(form, objIndex);
		FrmGetObjectBounds(form, objIndex, &r);
		r.extent.x = width;
		FrmSetObjectBounds(form, objIndex, &r);
		
		/* Set the remainder of the list parameters. */
		LstSetListChoices(lst, names, 1 + CurrentSource->schema.numFields);

		if ( FindInfo.type == DSF_FILTER_TYPE_SIMPLE &&
			 FindInfo.data.simple->field < CurrentSource->schema.numFields + 2)
			LstSetSelection(lst, FindInfo.data.simple->field);		
		else
			LstSetSelection(lst, 0);
			
		LstSetHeight(lst, 1 + CurrentSource->schema.numFields);
		
		ctl = GetObjectPtr(form, ctlID_FindDialog_FieldTrigger);
		CtlSetLabel(ctl, LstGetSelectionText(lst,LstGetSelection(lst)));
		

		FreeFindInfoData( FindInfo );
		FindInfo.data.any = 0;
		

        FrmDrawForm(form);
        FrmSetFocus(form, FrmGetObjectIndex(form, ctlID_FindDialog_Field));
	}
        return true;

/* unuse for a popup form*/
    case frmCloseEvent:
        if (names) {
            DeallocateMemPtr(names[0]);
            DeallocateMemPtr(names);
            names = 0;
        }
        form = FrmGetActiveForm();
        fld = GetObjectPtr(form, ctlID_FindDialog_Field);
        strH = FldGetTextHandle(fld);
        FldSetTextHandle(fld, 0);
        DeallocateMemHandle(strH);

        return false;
    case ctlSelectEvent:
        switch (event->data.ctlSelect.controlID) {
        case ctlID_FindDialog_FindButton: 


            form = FrmGetActiveForm();

            /*The field must be reseted to use strH*/
            fld = GetObjectPtr(form, ctlID_FindDialog_Field);
            strH = FldGetTextHandle(fld);
            FldSetTextHandle(fld, 0);

            str = MemHandleLock(strH);
            ctl = GetObjectPtr(form, ctlID_FindDialog_CaseSensCheckbox);
			
			
			FindInfo.type = DSF_FILTER_TYPE_SIMPLE;

#ifdef CONFIG_AFILTERS			
			if ( str[0] == '(' && CurrentSource->driver->class_ops.CompileScript ) {
				ScriptContext sc;
				CalculationData *calc;
				char *sctxt;
				UInt8 *bc;

				calc = AllocateMemPtr( sizeof(CalculationData) );
				MemSet( calc, sizeof(CalculationData), 0 );

				calc->scriptText = AllocateMemHandle( StrLen( str ) + 1 );
				sctxt = MemHandleLock( calc->scriptText );
				StrCopy( sctxt, str );
				MemHandleUnlock( calc->scriptText );

				sc.src = CurrentSource;
				sc.context = RECORD_FIELD_WRITING_CONTEXT;
				sc.flags = ScriptContext_DataAccess_RTC;
				sc.rec_field = 0xffff; // all fields accessible
				
				if (CurrentSource->driver->class_ops.CompileScript(&sc, calc, false )) {
					bc = MemHandleLock( calc->bytecode );
					
					MemHandleUnlock( calc->bytecode );
					FindInfo.type = DSF_FILTER_TYPE_SCRIPT;
					FindInfo.data.any = calc;
					
				}
				
				DeallocateMemHandle( calc->scriptText );
				

			}
#endif // if CONFIG_AFILTERS

			if ( FindInfo.type == DSF_FILTER_TYPE_SIMPLE ) {
				SimpleFilterData *sfd;
				
				sfd = AllocateMemPtr( sizeof( SimpleFilterData ) );
				FindInfo.data.simple = sfd;
				
				if ((FindInfo.data.simple->caseSensitive = CtlGetValue(ctl))) {
					StrNCopy(FindInfo.data.simple->matchString, str,32);
				}
				else {
					char tempo[32];
					StrNCopy(tempo, str, 32);
					tempo[31] = '\0';
					StrToLower(FindInfo.data.simple->matchString, tempo);
				}


				FindInfo.data.simple->matchString[31] = 0;
				MemPtrUnlock(str);
				
				ctl = GetObjectPtr(form, ctlID_FindDialog_WholeWordCheckbox);
				FindInfo.data.simple->wholeWord = CtlGetValue(ctl);
				
				lst = GetObjectPtr(form, ctlID_FindDialog_FieldList);
				FindInfo.data.simple->field = LstGetSelection(lst);
				
				/* Update the database flags with the current checkbox settings. */
				CurrentSource->ops.SetOption_Boolean(CurrentSource,
													 optionID_LFind_CaseSensitive,
													 FindInfo.data.simple->caseSensitive);
				CurrentSource->ops.SetOption_Boolean(CurrentSource,
													 optionID_LFind_WholeWord,
													 FindInfo.data.simple->wholeWord);
				

			} 



			if ( FindInfo.forFilterDefinition > -1 ) {
				DataSourceFeature_Filter filt;
				MemHandle h;

				filt.flags = FindInfo.type;
				fld = GetObjectPtr( form, ctlID_FindDialog_Name );
				h = FldGetTextHandle( fld );
				if ( h ) {
					StrCopy( filt.name, MemHandleLock( h ));
					MemHandleUnlock( h );
				}
				else filt.name[0] = '\0';

				if ( filt.name[0] == '\0' )
					StrCopy( filt.name, "Un-named" );

				filt.data.any = FindInfo.data.any;

				CurrentSource->ops.StoreFeature( CurrentSource,
												 featureFilter,
												 FeatureEditor_FeatureNum,
												 &filt );

				

				FrmUpdateForm( formID_NamesEditor, updateChangeForm );
			}
			else {

				
				if (FindInfo.type != DSF_FILTER_TYPE_SIMPLE ||
					FindInfo.data.simple->matchString[0] != '\0' ) {
				
					/* Do the actual find operation. */
					
					if (!(DBPrefs.flags & prefFlagUnuseFilters)
                        && CurrentSource->ops.Capable(CurrentSource, DS_CAPABLE_LISTVIEW_FILTER)) {
						UInt16 frec;

						lst = GetObjectPtr(form, ctlID_FindDialog_OperationList);
						
						if ( CurrentSource->ops.Search != 0 )
							frec = CurrentSource->ops.Search(CurrentSource, &FindInfo, dbTempResultsFilter,
															 ( LstGetSelection( lst ) == 1 ? FILTER_UNION : FILTER_INTERSECTION ));
						else
							frec = DS_Search(CurrentSource, &FindInfo, dbTempResultsFilter,
															 ( LstGetSelection( lst ) == 1 ? FILTER_UNION : FILTER_INTERSECTION ));
							
//						CurrentSource->currentcategory = trc;
						if (CurrentSource->ops.Capable( CurrentSource, DS_CAPABLE_LISTVIEW_FILTER)) {
							
						CurrentSource->ops.SetFilter( CurrentSource, dbTempResultsFilter, false, &FindInfo );
						if ( LastFormId[0] == formID_ListView )
							ListViewUpdateAfterFind( dbTempResultsFilter, frec, &FindInfo );
						}
						else {
						if ( LastFormId[0] == formID_ListView )
							ListViewUpdateAfterFind( dbNoFilter, frec, &FindInfo );
						}							
					} else {
						UInt16 frec;

						FindInfo.lastFound = 0xFFFF;
						if ( CurrentSource->ops.Search != 0 )
							frec = CurrentSource->ops.Search(CurrentSource, &FindInfo, dbNoFilter, FILTER_INTERSECTION);
						else 
							frec = DS_Search(CurrentSource, &FindInfo, dbNoFilter, FILTER_INTERSECTION);
						// Just set the top index, we can't filter.
						if ( LastFormId[0] == formID_ListView )
							ListViewUpdateAfterFind( dbNoFilter, frec, &FindInfo );

					}
				}
			}

            form = FrmGetActiveForm();
            /* Remove the find dialog from the screen. */
            if (names) {
                DeallocateMemPtr(names[0]);
                DeallocateMemPtr(names);
                names = 0;
            }
            DeallocateMemHandle(strH);
            FrmReturnToForm(0);

            return true;

        case ctlID_FindDialog_CancelButton:
		{
			SimpleFilterData *sfd;
			
			sfd = AllocateMemPtr( sizeof( SimpleFilterData ) );
			FindInfo.data.simple = sfd;
			FindInfo.data.simple->matchString[0] = '\0';

            form = FrmGetActiveForm();
            /* Remove the find dialog from the screen. */
            if (names) {
                DeallocateMemPtr(names[0]);
                DeallocateMemPtr(names);
                names = 0;
            }
            fld = GetObjectPtr(form, ctlID_FindDialog_Field);
            strH = FldGetTextHandle(fld);
            FldSetTextHandle(fld, 0);
            DeallocateMemHandle(strH);
            FrmReturnToForm(0);
            return true;
		}
        }
        break;

    default:
        break;
    }

    return false;
}


#endif

Boolean
RecordPassesSimpleFilter( DataSource *source, FindData *findInfo, UInt16 recNum ) {

    DataSourceGetter getter;
    void * rec_data;
    Boolean match;


	/* Lock down the record. */
	source->ops.LockRecord(source, recNum, false,
						   &getter, &rec_data);
	
	match = false;
	if (findInfo->data.simple->field == 0) {
		UInt16 i;
		
		/* Search all string fields for the string. */
		for (i = 0; i < source->schema.numFields; ++i) {
			/* Compare the string to the field. */
			match = CmpField(source, &getter, rec_data, i,
							 findInfo->data.simple->matchString, 
							 findInfo->data.simple->caseSensitive,
							 findInfo->data.simple->wholeWord );
			
			/* Bail out if we found a match. */
			if (match) break;
		}
	} else {
		/* Search the single field. */
		match = CmpField(source, &getter, rec_data, 
						 findInfo->data.simple->field - 1, 
						 findInfo->data.simple->matchString,
						 findInfo->data.simple->caseSensitive,
						 findInfo->data.simple->wholeWord);
	}


	/* Unlock the record. */
	source->ops.UnlockRecord(source, recNum, false, rec_data);
	
	return match;



}


Boolean 
RecordPassesScriptFilter( DataSource *source, FindData *findInfo, UInt16 recNum ) {

    DataSourceGetter getter;
    void * rec_data;
    Boolean match;
	CalculatedValue cv;
	ScriptContext sc;
	UInt8 *bc;


	/* Lock down the record. */
	source->ops.LockRecord(source, recNum, false, &getter, &rec_data);

	bc = MemHandleLock( findInfo->data.script->bytecode );
	sc.flags = ScriptContext_DataAccess_EXE;
	sc.src = CurrentSource;
	sc.data_access.ra.getter = &getter;
	sc.data_access.ra.rec_data = rec_data;
	sc.rec_field = recNum;


	source->ops.ExecuteScript( &sc, bc, &cv );

	match = (cv.value.i != 0 );
	

	MemHandleUnlock( findInfo->data.script->bytecode );
//	DeallocateMemHandle( calc.bytecode );
//	DeallocateMemHandle( calc.scriptText );
	
	/* Unlock the record. */
	source->ops.UnlockRecord(source, recNum, false, rec_data);
	
	return match;



}







