#include <PalmOS.h>
#include "db.h"
#include "double.h"
#include "calcvalue.h"
#include "report.h"

void printReport( DataSource *source, UInt16 field, char *buff ) {
	void *packed;
	DataSourceGetter getter;
	UInt16 j;
	char *s;

	switch( source->schema.fields[field].type ) {
	case FIELD_TYPE_INTEGER:
	{
		Int32 c,min,max,n;
		double sum;
		Int16 above,below,eq;

		min = 0x7fffffff;
		max = 0xffffffff;
		
		j = 0;
		above = below = eq = 0;
		n = 0;
		sum = 0.0;

		while( source->ops.Seek( source, &j, 0, dmSeekForward ) ) {
			source->ops.LockRecord(source, j, false, 
								   &getter, &packed);
			c = getter.GetInteger(packed, field );
			if ( c > max ) max = c;
			if ( c < min ) min = c;
			if ( c > 0 ) above++;
			else if ( c < 0 ) below++;
			else eq++;

			sum += DoubleIToD( c );
			source->ops.UnlockRecord(source, j, false, packed);
			j++;
			n++; // can't use j since it might not always increase by 1
		}
		StrPrintF( buff, "\tThis is an Integer\n\nMax: %ld\nMin: %ld\n(+,-,0): (%d,%d,%d)\nSum: ", max, min, above, below, eq );
		s = buff;
		while ( *s ) s++;
		DoublePrintValue( sum, s, 2 );
		while( *s ) s++;
		StrCopy( s, "\nAvg: " );
		s += 6;
		sum /= n;
		DoublePrintValue( sum, s, 2 );
		break;
	
	}
	case FIELD_TYPE_FLOAT:
	{
		Int32 n;
		double sum, min, max, c;
		Int16 above,below,eq;

		
		j = 0;
		above = below = eq = 0;
		n = 0;
		sum = 0.0;
		min = 0.0;
		max = 0.0;
		while( source->ops.Seek( source, &j, 0, dmSeekForward ) ) {
			source->ops.LockRecord(source, j, false, 
								   &getter, &packed);

			c = getter.GetFloat(packed, field );
			if ( n == 0 ) {
				min = c;
				max = c;
			} else {
				if ( c > max ) max = c;
				if ( c < min ) min = c;
			}
			if ( c > 0 ) above++;
			else if ( c < 0 ) below++;
			else eq++;

			sum += c;
			source->ops.UnlockRecord(source, j, false, packed);
			j++;
			n++; // can't use j since it might not always increase by 1
		}
		StrPrintF( buff, "\tThis is a Float\n\nMax: ");
		s = buff;
		while ( *s ) s++;

		DoublePrintValue( max, s, 2 );
		while( *s ) s++;
		StrCopy( s, "\nMin:  " );
		while( *s ) s++;
		DoublePrintValue( min, s, 2 );		
		while( *s ) s++;

		StrPrintF( s, "\n(+,-,0): (%d,%d,%d)\n", above, below, eq );
		while( *s ) s++;
		StrCopy( s, "Sum: " );
		while( *s ) s++;

		DoublePrintValue( sum, s, 2 );		
		while( *s ) s++;

		StrCopy( s, "\nAvg: " );
		while( *s ) s++;		
		sum /= n;
		DoublePrintValue( sum, s, 2 );

		break;
	}
	case FIELD_TYPE_BOOLEAN:
	{
		Int32 n;
		Int16 on;
		Boolean c;
		
		j = 0;
		n = 0;
		on = 0;

		while( source->ops.Seek( source, &j, 0, dmSeekForward ) ) {
			source->ops.LockRecord(source, j, false, 
								   &getter, &packed);

			c = getter.GetBoolean(packed, field );
			if ( c )  on++;

			source->ops.UnlockRecord(source, j, false, packed);
			j++;
			n++; // can't use j since it might not always increase by 1
		}
		StrPrintF( buff, "\tThis is a Boolean\n\nOn: %d\nOff: %ld", on, n-on);
		break;
	}	
	case FIELD_TYPE_LIST:
	{
		Int32 n;
		Int16 *count;		
		Int16 max,nf = source->schema.fields[field].data.list.numItems;
		UInt8 c;


		count = MemPtrNew( sizeof( Int16 ) * (nf+1) );
		
		for( j = 0; j <= nf; j++ ) count[j] = 0;
		j = 0;
		
		while( source->ops.Seek( source, &j, 0, dmSeekForward ) ) {
			source->ops.LockRecord(source, j, false, 
								   &getter, &packed);
			
			c = getter.GetListItem(packed, field );
			if ( c < nf ) count[c]++;
			else count[nf]++;

			source->ops.UnlockRecord(source, j, false, packed);
			j++;
			n++; // can't use j since it might not always increase by 1
		}
		StrPrintF( buff, "\tThis is a List\n\n");
		s = buff;
		while ( *s ) s++;
		
		// cap it off at 5 lines...
		max = nf;
		if (max > 4) max = 4;
						
		for( j = 0; j < max; j++ ) {
			StrPrintF( s, " - %s: %d\n",
					   source->schema.fields[field].data.list.items[j],
					   count[j]);
			s = buff;
			while ( *s ) s++;
		}
		for( ; j < nf; j++ ) {
			count[nf] += count[j];
		}
		StrPrintF( s, " - <Other>: %d", count[nf]);
		
		MemPtrFree( count );
		break;
	}	
	case FIELD_TYPE_CALCULATED:
	{
		
		StrPrintF( buff, "\tThis is a Calculated Field\n\nReturn Type: " );
		s = buff + 41;

		switch( source->schema.fields[field].data.calc.returnType ) {
		case FIELD_TYPE_INTEGER:
			StrPrintF( s, "Integer" ); break;
		case FIELD_TYPE_FLOAT:
			StrPrintF( s, "Float" ); break;
		case FIELD_TYPE_STRING:
			StrPrintF( s, "String" ); break;
		case CV_FIELD_TYPE_UNKNOWN:
			StrPrintF( s, "Unknown, or Indeterminate" ); break;
		case FIELD_TYPE_DATE:
			StrPrintF( s, "Date" ); break;
		case FIELD_TYPE_TIME:
			StrPrintF( s, "Time" ); break;
		case CV_FIELD_TYPE_DURATION:
			StrPrintF( s, "Duration" ); break;
		case CV_FIELD_TYPE_BROKEN:
			StrPrintF( s, "This script is broken" ); break;
			
		}
		while( *s != '\0' ) s++;

		StrPrintF( s, "\nAccess Requirements:" );
		s += 21;

		if ( source->schema.fields[field].data.calc.access == NO_EXTERNAL_ACCESS ) {
			StrPrintF( s, "\n - self contained" );
			s += 18;
		}
		else {
			if ( ScriptRequiresAtLeast( &source->schema.fields[field].data.calc, SINGLE_RECORD_READING ) ) {
			StrPrintF( s, "\n - record reading" );
			s += 18;
			}
			if ( ScriptRequiresAtLeast( &source->schema.fields[field].data.calc,  FIELD_READING ) ) {
			StrPrintF( s, "\n - field reading" );
			s += 17;
			}
			if ( ScriptRequiresAtLeast( &source->schema.fields[field].data.calc,  MULTI_RECORD_READING ) ) {
			StrPrintF( s, "\n - multi record reading" );
			s += 24;
			}
			if ( source->schema.fields[field].data.calc.access & ARRANGING ) {
			StrPrintF( s, "\n - arranging" );
			s += 13;
			}
		}

		if ( source->schema.fields[field].data.calc.returnType == 
			 CV_FIELD_TYPE_INTEGER ||
			 source->schema.fields[field].data.calc.returnType == 
			 CV_FIELD_TYPE_FLOAT  ||
			 source->schema.fields[field].data.calc.returnType == 
			 CV_FIELD_TYPE_DURATION  
			) {
			
			Int32 n;
			double sum, min, max, c;
			Int16 above,below,eq;
			CalculatedValue cv;
			
			j = 0;
			above = below = eq = 0;
			n = 0;
			sum = 0.0;
			min = 0.0;
			max = 0.0;
			while( source->ops.Seek( source, &j, 0, dmSeekForward ) ) {
				source->ops.LockRecord(source, j, false, 
									   &getter, &packed);
				
				
				getter.GetCalculated(packed, field, &cv );
				if ( cv.info == CV_FIELD_TYPE_FLOAT ) c = cv.value.d;
				else c = DoubleIToD( cv.value.i );
				
				if ( n == 0 ) {
					min = c;
					max = c;
				} else {
					if ( c > max ) max = c;
					if ( c < min ) min = c;
				}
				if ( c > 0 ) above++;
				else if ( c < 0 ) below++;
				else eq++;
				
				sum += c;
				source->ops.UnlockRecord(source, j, false, packed);
				j++;
				n++; // can't use j since it might not always increase by 1
			}
			
			StrCopy( s, "\nMax:  " );
			s += 7;
			DoublePrintValue( max, s, 2 );
			while( *s ) s++;
			StrCopy( s, "\nMin:  " );
			s += 7;
			DoublePrintValue( min, s, 2 );		
			while( *s ) s++;
			
			StrPrintF( s, "\n(+,-,0): (%d,%d,%d)\n", above, below, eq );
			while( *s ) s++;
			StrCopy( s, "Sum: " );
			s += 5;
			
			DoublePrintValue( sum, s, 2 );		
			while( *s ) s++;
			
			StrCopy( s, "\nAvg: " );
			s += 6;
			sum /= n;
			DoublePrintValue( sum, s, 2 );
		}
		
		break;
	}	

	default:				   
		StrCopy( buff, "No report is available for this field type..." );
		break;
	}	

}


